import json
from typing import List, Optional, Callable
from .base import BaseCLIWrapper

class GeminiCLIWrapper(BaseCLIWrapper):
    """
    Wrapper for the Gemini CLI tool using stream-json output.
    """

    def __init__(self, command: str = "gemini", auto_approve: bool = True, model: Optional[str] = None, work_dir: Optional[str] = None, read_only: bool = False, sandbox_manager=None):
        super().__init__(command, auto_approve, model, work_dir, read_only=read_only, sandbox_manager=sandbox_manager)

    def _construct_command(self, prompt_text: str) -> List[str]:
        cmd_args = self.command.split() + ["-o", "stream-json"]
        
        if self.auto_approve and not self.read_only:
            cmd_args.append("--yolo")
        
        if self.model:
            cmd_args.extend(["--model", self.model])
        
        if self.session_id:
            cmd_args.extend(["--resume", self.session_id])
            
        cmd_args.append(prompt_text)
        return cmd_args

    def _process_stream_line(self, line: str, source: str, callback: Callable[[str, str], None]) -> None:
        if source == "stderr":
            callback(line, source)
            return

        # For stdout, parse JSON
        line = line.strip()
        if not line:
            return

        try:
            if not line.startswith("{"):
                return
            data = json.loads(line)
            if data.get("type") == "message" and data.get("role") == "assistant":
                content = data.get("content", "")
                if content:
                    callback(content, source)
        except json.JSONDecodeError:
            pass

    def _parse_output(self, stdout: str) -> str:
        output_lines = stdout.splitlines()
        response_content = []
        
        for line in output_lines:
            line = line.strip()
            if not line:
                continue
                
            try:
                if not line.startswith("{"):
                    continue
                    
                data = json.loads(line)
                
                # Capture session ID
                if data.get("type") == "init" and "session_id" in data:
                    if not self.session_id:
                        self.session_id = data["session_id"]
                
                # Collect assistant messages
                if data.get("type") == "message" and data.get("role") == "assistant":
                    content = data.get("content", "")
                    if content:
                        response_content.append(content)
                        
            except json.JSONDecodeError:
                continue

        return "".join(response_content).strip()



from textual.widgets import Static, RichLog
from textual.app import ComposeResult
from textual.reactive import reactive
from rich.panel import Panel
from rich.table import Table

class TimelineWidget(Static):
    """Displays the workflow timeline/steps."""

    current_step_index = reactive(0)
    steps = reactive([])

    def __init__(self, steps: list, **kwargs):
        super().__init__(**kwargs)
        self.steps = steps

    def on_mount(self):
        self.update_timeline()

    def watch_current_step_index(self, old_val, new_val):
        self.update_timeline()

    def update_timeline(self):
        grid = Table.grid(expand=True)
        grid.add_column(style="dim")
        grid.add_column()
        grid.add_column(justify="right")

        for i, step in enumerate(self.steps):
            idx = i + 1
            name = step.name

            if i < self.current_step_index:
                status = "[green]✓[/green]"
                style = "dim"
            elif i == self.current_step_index:
                status = "[blue]▶[/blue]"
                style = "bold blue"
            else:
                status = " "
                style = "dim"

            grid.add_row(f"{idx}.", f"[{style}]{name}[/{style}]", status)

        self.update(Panel(grid, title="Timeline", border_style="blue"))


class StatsWidget(Static):
    """Displays execution statistics."""

    execution_time = reactive("00:00")
    status_message = reactive("Ready")

    def render(self):
        grid = Table.grid(expand=True)
        grid.add_column(justify="left", ratio=1)
        grid.add_column(justify="right", ratio=1)

        grid.add_row(f"Status: {self.status_message}", f"Time: {self.execution_time}")
        return Panel(grid, title="Stats", border_style="green")



import os
import json
from typing import Dict, Any, Optional

class StateManager:
    def __init__(self, task_id: str, state_directory: str):
        self.task_id = task_id
        self.state_directory = state_directory
        self.task_dir = os.path.join(self.state_directory, self.task_id)
        self.state_file = os.path.join(self.task_dir, f"task_{self.task_id}.json")
        self.state: Dict[str, Any] = {
            "current_step_index": 0,
            "context": {},
            "timeline": []
        }

        os.makedirs(self.task_dir, exist_ok=True)
        self.load_state()

    def load_state(self):
        if os.path.exists(self.state_file):
            try:
                with open(self.state_file, 'r') as f:
                    self.state = json.load(f)
                    if "timeline" not in self.state:
                        self.state["timeline"] = []
            except Exception as e:
                print(f"Error loading state from {self.state_file}: {e}. Starting fresh.")
                self.state = {"current_step_index": 0, "context": {}, "timeline": []}

    def save_state(self):
        with open(self.state_file, 'w') as f:
            json.dump(self.state, f, indent=2)

    def write_context_to_file(self, step_index: int, step_name: str, key: str, content: str) -> str:
        """Writes context content to a formatted MD file inside task directory and returns the relative path."""
        safe_step_name = step_name.lower().replace(" ", "_")
        filename = f"step_{step_index}_{safe_step_name}_{key}.md"
        filepath = os.path.join(self.task_dir, filename)

        # Check if content is JSON and wrap it in a code block if so
        formatted_content = content
        try:
            parsed = json.loads(content)
            if isinstance(parsed, (dict, list)):
                formatted_content = f"```json\n{content}\n```"
        except (json.JSONDecodeError, TypeError):
            pass

        with open(filepath, 'w', encoding='utf-8') as f:
            f.write(formatted_content)

        return filepath

    def read_context_from_file(self, path: str) -> str:
        """Reads context content from a file."""
        if not os.path.isabs(path):
            if os.path.exists(path):
                target_path = path
            elif os.path.exists(os.path.join(self.task_dir, os.path.basename(path))):
                # Try task directory
                target_path = os.path.join(self.task_dir, os.path.basename(path))
            elif os.path.exists(os.path.join(self.state_directory, os.path.basename(path))):
                # Fallback to state directory (backward compatibility)
                target_path = os.path.join(self.state_directory, os.path.basename(path))
            else:
                return f"[Error: Context file {path} not found]"
        else:
             target_path = path

        try:
            with open(target_path, 'r', encoding='utf-8') as f:
                return f.read()
        except Exception as e:
            return f"[Error reading context file: {e}]"

    def get_context(self, key: str) -> Any:
        return self.state["context"].get(key)

    def set_context(self, key: str, value: Any):
        self.state["context"][key] = value

    def remove_context(self, key: str):
        if key in self.state["context"]:
            del self.state["context"][key]

    def get_current_step_index(self) -> int:
        return self.state["current_step_index"]

    def set_current_step_index(self, index: int):
        self.state["current_step_index"] = index

    def delete_file(self, filepath: str) -> bool:
        """Deletes a file if it exists. Returns True if deleted, False otherwise."""
        if filepath and os.path.exists(filepath):
            os.remove(filepath)
            return True
        return False

    def add_timeline_entry(self, entry: Dict[str, Any]):
        """Adds an entry to the execution timeline."""
        if "timeline" not in self.state:
            self.state["timeline"] = []
        self.state["timeline"].append(entry)

    def get_timeline(self) -> list:
        """Returns the execution timeline."""
        return self.state.get("timeline", [])



from typing import List, Optional, Callable
from .base import BaseCLIWrapper
from .utils import strip_ansi_codes

class GitHubCopilotCLIWrapper(BaseCLIWrapper):
    """
    Wrapper for the GitHub Copilot CLI (gh copilot).
    Uses 'gh copilot explain' for interactions.
    """

    def __init__(self, command: str = "gh", auto_approve: bool = True, model: Optional[str] = None, work_dir: Optional[str] = None, read_only: bool = False, sandbox_manager=None):
        """
        Args:
            command (str): The base command (defaults to 'gh').
            auto_approve (bool): Whether to auto-approve tool calls (not used for Copilot currently).
            model (str, optional): Not typically used for Copilot CLI, but kept for consistency.
            work_dir (str, optional): The working directory.
            read_only (bool): Whether to run in read-only mode (default: False).
            sandbox_manager (SandboxManager, optional): Manager for Docker sandbox execution.
        """
        super().__init__(command, auto_approve, model, work_dir, read_only=read_only, sandbox_manager=sandbox_manager)

    def _construct_command(self, prompt_text: str) -> List[str]:
        # 'gh copilot explain' is used for general queries and code explanations
        # It handles natural language prompts well.
        cmd_args = self.command.split() + ["copilot", "explain", prompt_text]
        return cmd_args

    def _process_stream_line(self, line: str, source: str, callback: Callable[[str, str], None]) -> None:
        # Strip ANSI codes from streaming output
        clean_line = strip_ansi_codes(line)
        callback(clean_line, source)

    def _parse_output(self, stdout: str) -> str:
        # Strip ANSI codes from final output
        return strip_ansi_codes(stdout).strip()



from textual.app import App
from typing import List, Optional
from ..ui import UserInterface

class TuiUIAdapter(UserInterface):
    """
    Adapter that bridges the Workflow engine (UserInterface protocol)
    to the Textual App instance.
    """
    def __init__(self, app: App):
        self.app = app

    def print_step_header(self, step_number: int, step_name: str, work_dir: str = None):
        self.app.call_from_thread(self.app.on_step_start, step_number, step_name)

    def print_assignment(self, role: str, tool: str, model: str):
        self.app.call_from_thread(self.app.log, f"[bold]Assignment:[/bold] Role={role}, Tool={tool}, Model={model}")

    def print_timeout(self, timeout_val: int):
        self.app.call_from_thread(self.app.log, f"[yellow]Timeout set to {timeout_val}s[/yellow]")

    def print_input_loading(self, input_key: str, path: str):
        self.app.call_from_thread(self.app.log, f"Loading input {input_key} from {path}")

    def print_tool_start(self, step_name: str, tool_name: str):
        self.app.call_from_thread(self.app.log, f"[green]Starting {step_name} via {tool_name}...[/green]")

    def start_spinner(self, message: str):
        # Textual handles loading via widgets or headers mostly,
        # but we can update status
        self.app.call_from_thread(self.app.update_status, message)

    def stop_spinner(self):
        self.app.call_from_thread(self.app.update_status, "Ready")

    def print_output(self, step_name: str, output: dict):
        self.app.call_from_thread(self.app.log, f"[bold success]--- Output from {step_name} ---[/bold success]")
        import json
        self.app.call_from_thread(self.app.log, json.dumps(output, indent=2))

    def print_error(self, message: str):
        self.app.call_from_thread(self.app.log, f"[bold red]ERROR: {message}[/bold red]")

    def print_warning(self, message: str):
        self.app.call_from_thread(self.app.log, f"[bold yellow]WARNING: {message}[/bold yellow]")

    def print_info(self, message: str):
        self.app.call_from_thread(self.app.log, f"[dim]{message}[/dim]")

    def print_success(self, message: str):
        self.app.call_from_thread(self.app.log, f"[bold green]{message}[/bold green]")

    def print_stream_line(self, line: str, source: str = "stdout"):
        # For streaming, we append to the log widget
        self.app.call_from_thread(self.app.stream_log, line)

    def input(self, prompt: str) -> str:
        # Input in TUI is tricky if the workflow is in a thread.
        # We'd need to pause execution and wait for UI event.
        # For now, we'll raise NotImplementedError or handle it if we have time to implement a modal.
        # Given the complexity, we might skip full interactive input support in the first pass
        # or implement a simple workaround.
        # TODO: Implement interactive input via Worker.wait() or similar
        return ""

    def print_user_intervention(self, step_name: str, files: list):
        self.app.call_from_thread(self.app.log, f"[bold yellow]USER INTERVENTION REQUIRED for {step_name}[/bold yellow]")
        for k, p in files:
            self.app.call_from_thread(self.app.log, f" - {k}: {p}")



from textual.app import App, ComposeResult
from textual.widgets import Header, Footer, RichLog
from textual.containers import Container, Horizontal, Vertical
from textual.worker import Worker
from datetime import datetime
import asyncio

from .widgets import TimelineWidget, StatsWidget
from .adapter import TuiUIAdapter
from ..workflow import SoftwareCompanyWorkflow

class SoftwareCompanyApp(App):
    CSS = """
    TimelineWidget {
        width: 30%;
        height: 100%;
        dock: left;
    }

    StatsWidget {
        height: 5;
        dock: bottom;
    }

    RichLog {
        border: solid green;
    }
    """

    def __init__(self, workflow_factory, **kwargs):
        super().__init__(**kwargs)
        self._workflow_factory = workflow_factory
        self.workflow = None
        self.start_time = None

    def compose(self) -> ComposeResult:
        yield Header()
        yield Horizontal(
            TimelineWidget(steps=[]), # Initial empty, will update
            RichLog(highlight=True, markup=True, id="main_log"),
        )
        yield StatsWidget()
        yield Footer()

    async def on_mount(self):
        # We need to instantiate the workflow here because we need to pass the adapter
        # which needs 'self' (the app).

        adapter = TuiUIAdapter(self)

        # The factory is a partial or callable that takes 'ui'
        self.workflow = self._workflow_factory(ui=adapter)

        # Initialize widgets
        timeline = self.query_one(TimelineWidget)
        timeline.steps = self.workflow.steps
        timeline.current_step_index = self.workflow.state_manager.get_current_step_index()

        self.start_time = datetime.now()
        self.set_interval(1, self.update_timer)

        # Run the workflow in a worker
        self.run_worker(self.run_workflow_task, exclusive=True)

    async def run_workflow_task(self):
        try:
            # We assume 'run' is an async method on the workflow
            # We need to know arguments for run().
            # Ideally the factory or main.py set up the prompt/args.
            # But workflow.run() takes optional args.
            # We might need to pass them in via factory or similar.
            # Let's assume the workflow object is fully configured.

            # The run method signature is async def run(self, user_request: str = "", single_step: bool = False):
            # We might need to pass user_request if it's a new task.
            # However, main.py logic handles initializing context with user_request BEFORE calling run if it was passed.
            # Let's check main.py again...
            # Yes, main.py calls `await workflow.run(prompt if prompt else "", single_step=args.step)`
            # So we need to be able to pass these.

            # We will rely on the factory to have set up the workflow instance correctly,
            # or simply call run with empty string if context is already set.
            # But wait, main.py logic is:
            # `workflow = SoftwareCompanyWorkflow(...)`
            # `await workflow.run(...)`

            # So `self.workflow` is the instance.

            # Hack: we will check if "user_request" is in context. If not, we might need it.
            # But the TuiAdapter injection happens at init.

            await self.workflow.run()

        except Exception as e:
            self.log(f"[bold red]Workflow Crashed: {e}[/bold red]")

    def on_step_start(self, step_number: int, step_name: str):
        timeline = self.query_one(TimelineWidget)
        # step_number is 1-based index
        timeline.current_step_index = step_number - 1
        self.log(f"[bold blue]Step {step_number}: {step_name}[/bold blue]")

    def log(self, message: str):
        log_widget = self.query_one("#main_log", RichLog)
        log_widget.write(message)

    def stream_log(self, text: str):
        log_widget = self.query_one("#main_log", RichLog)
        # Using write with end="" if possible, but RichLog.write adds newline by default?
        # RichLog writes lines. Streaming might be tricky if we want char-by-char.
        # But 'text' from callbacks is usually chunks.
        log_widget.write(text)

    def update_status(self, message: str):
        stats = self.query_one(StatsWidget)
        stats.status_message = message

    def update_timer(self):
        if self.start_time:
            delta = datetime.now() - self.start_time
            stats = self.query_one(StatsWidget)
            stats.execution_time = str(delta).split('.')[0]



import os
import subprocess
import time
from typing import Dict, Any, List, Optional

class SandboxManager:
    """
    Manages the lifecycle of a Docker container for sandboxed execution.
    """

    def __init__(self, config: Dict[str, Any], task_id: str, work_dir: str):
        self.config = config.get("sandbox", {})
        self.task_id = task_id
        self.work_dir = os.path.abspath(work_dir)
        self.enabled = self.config.get("enabled", False)
        self.image = self.config.get("image", "python:3.11-slim")
        self.mount_path = self.config.get("mount_path", "/workspace")
        self.container_name = f"autogen_sandbox_{self.task_id}"

    def is_active(self) -> bool:
        """Returns true if sandboxing is enabled."""
        return self.enabled

    def start(self) -> None:
        """
        Starts the Docker container in detached mode with volume mounts.
        Idempotent: if container already exists and is running, does nothing.
        If it exists but is stopped, restarts it.
        """
        if not self.enabled:
            return

        # Check if container exists
        status = self._get_container_status()

        if status == "running":
            return

        if status == "exited" or status == "created":
            self._run_cmd(["docker", "start", self.container_name])
            return

        # Clean up if it exists in some other state (e.g. paused/dead?)
        # For now, if it's not running/exited/created, we assume we need to create it.
        # But to be safe, we can try to remove it first if it exists.
        if status:
             self._run_cmd(["docker", "rm", "-f", self.container_name])

        uid = os.getuid()
        gid = os.getgid()

        cmd = [
            "docker", "run", "-d", "--rm",
            "--name", self.container_name,
            "-v", f"{self.work_dir}:{self.mount_path}",
            "-w", self.mount_path,
            "-u", f"{uid}:{gid}",
            self.image,
            "tail", "-f", "/dev/null"
        ]

        print(f"[Sandbox] Starting container {self.container_name}...")
        self._run_cmd(cmd)

        # Wait a moment for it to be ready?
        time.sleep(1)

    def stop(self) -> None:
        """Stops and removes the Docker container."""
        if not self.enabled:
            return

        status = self._get_container_status()
        if status:
            print(f"[Sandbox] Stopping container {self.container_name}...")
            # We used --rm, so stopping it should remove it.
            # But using -f to be sure.
            self._run_cmd(["docker", "rm", "-f", self.container_name])

    def wrap_command(self, command: List[str], work_dir: Optional[str] = None) -> List[str]:
        """
        Prefixes the command with 'docker exec ...' logic.

        Args:
            command: The command and arguments as a list.
            work_dir: Optional working directory INSIDE the container.
                      If None, uses the mount_path.
        """
        if not self.enabled:
            return command

        # We need to map the host work_dir to the container path if provided
        # If the provided work_dir is absolute path on host, we need to convert it to container path
        # Assumption: work_dir is a subdirectory of self.work_dir (host project root)

        container_cwd = self.mount_path
        if work_dir:
            # Normalize paths
            host_root = self.work_dir
            target_path = os.path.abspath(work_dir)

            if target_path.startswith(host_root):
                rel_path = os.path.relpath(target_path, host_root)
                if rel_path != ".":
                    container_cwd = f"{self.mount_path}/{rel_path}".replace("\\", "/")

        exec_cmd = [
            "docker", "exec", "-i",
            "-w", container_cwd,
            self.container_name
        ]

        return exec_cmd + command

    def _get_container_status(self) -> Optional[str]:
        """Returns the status of the container or None if not found."""
        try:
            result = subprocess.run(
                ["docker", "inspect", "--format", "{{.State.Status}}", self.container_name],
                capture_output=True, text=True, check=False
            )
            if result.returncode == 0:
                return result.stdout.strip()
            return None
        except FileNotFoundError:
             # Docker not installed or not in PATH
             print("[Sandbox] Warning: Docker command not found.")
             return None

    def _run_cmd(self, cmd: List[str]) -> None:
        try:
            subprocess.run(cmd, check=True, capture_output=True)
        except subprocess.CalledProcessError as e:
            print(f"[Sandbox] Error running docker command: {e}")
            if e.stderr:
                print(f"[Sandbox] Stderr: {e.stderr.decode()}")
            raise
